# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the TransactionInstruction of the solana module."""

from typing import List, NamedTuple

from aea_ledger_solana.account import AccountMeta
from solders import instruction
from solders.pubkey import Pubkey as PublicKey


class TransactionInstruction(NamedTuple):
    """Transaction Instruction class."""

    keys: List[AccountMeta]
    """Public keys to include in this transaction Boolean represents whether this
    pubkey needs to sign the transaction.
    """
    program_id: PublicKey
    """Program Id to execute."""
    data: bytes = bytes(0)
    """Program input."""

    @classmethod
    def from_solders(cls, ixn: instruction.Instruction) -> "TransactionInstruction":
        """
        Convert from a `solders` instruction.

        :param ixn: The `solders` instruction.
        :param The `solana-py` instruction.
        :return: The `solders` instruction.
        """
        keys = [AccountMeta.from_solders(am) for am in ixn.accounts]
        program_id = PublicKey.from_bytes(bytes(ixn.program_id))
        return cls(keys=keys, program_id=program_id, data=ixn.data)

    def to_solders(self) -> instruction.Instruction:
        """
        Convert to a `solders` instruction.

        :return: The `solders` instruction.
        """
        accounts = [key for key in self.keys]
        return instruction.Instruction(
            program_id=self.program_id, data=self.data, accounts=accounts
        )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the Crypto implementation for the solana ledger."""

import base64
import hashlib
import json
from pathlib import Path
from typing import Optional, Union

import base58
from aea_ledger_solana.constants import _SOLANA
from cryptography.fernet import Fernet  # type: ignore
from solana.transaction import Transaction as SolanaTransaction
from solders.hash import Hash
from solders.instruction import Instruction
from solders.keypair import Keypair
from solders.message import MessageV0
from solders.pubkey import Pubkey
from solders.transaction import Transaction, VersionedTransaction

from aea.common import JSONLike
from aea.crypto.base import Crypto
from aea.crypto.helpers import DecryptError, KeyIsIncorrect


class SolanaCrypto(Crypto[Keypair]):
    """Class wrapping the Account Generation from Solana ledger."""

    identifier = _SOLANA

    def __init__(
        self,
        private_key_path: Optional[str] = None,
        password: Optional[str] = None,
        extra_entropy: Union[str, bytes, int] = "",
    ) -> None:
        """
        Instantiate a solana crypto object.

        :param private_key_path: the private key path of the agent
        :param password: the password to encrypt/decrypt the private key.
        :param extra_entropy: add extra randomness to whatever randomness from OS.
        """
        super().__init__(
            private_key_path=private_key_path,
            password=password,
            extra_entropy=extra_entropy,
        )

    @property
    def pubkey(self) -> Pubkey:
        """Pubkey object."""
        return self.entity.pubkey()

    @property
    def private_key(self) -> str:
        """
        Return a private key.

        64 random hex characters (i.e. 32 bytes) prefix.

        :return: a private key string in hex format
        """
        seed = bytes(self.entity.secret())
        private_key = seed + bytes.fromhex(self.public_key)
        return base58.b58encode(private_key).decode()

    @property
    def public_key(self) -> str:
        """
        Return a public key in hex format.

        :return: a public key string in hex format
        """
        return bytes(self.entity.pubkey()).hex()

    @property
    def address(self) -> str:
        """
        Return the address for the key pair.

        :return: an address string in hex format
        """
        return str(self.entity.pubkey())

    @classmethod
    def load_private_key_from_path(
        cls, file_name: str, password: Optional[str] = None
    ) -> Keypair:
        """
        Load a private key in base58 or bytes format from a file.

        :param file_name: the path to the hex file.
        :param password: the password to encrypt/decrypt the private key.
        :return: the Entity.
        """
        key_path = Path(file_name)
        private_key = cls.load(file_name, password)
        try:
            key = Keypair.from_base58_string(private_key)
        except Exception as e:
            raise KeyIsIncorrect(
                f"Error on key `{key_path}` load! : Error: {repr(e)}.{'try password?' if password is None else ''}"
            ) from e

        return key

    def sign_message(self, message: bytes, is_deprecated_mode: bool = False) -> str:
        """
        Sign a message in bytes string form.

        :param message: the message to be signed
        :param is_deprecated_mode: if the deprecated signing is used
        :return: signature of the message in string form
        """
        if is_deprecated_mode:
            raise ValueError("is_deprecated_mode is not supported at the moment")
        keypair = Keypair.from_base58_string(self.private_key)
        signed_msg = keypair.sign_message(message)
        return str(signed_msg)

    def sign_transaction(
        self, transaction: JSONLike, signers: Optional[list] = None
    ) -> JSONLike:
        """
        Sign a transaction in bytes string form.

        :param transaction: the transaction to be signed
        :param signers: list of signers
        :return: signed transaction
        """
        if "ixs" in transaction:
            return self._sign_ixs(ix_container=transaction)
        return self._sign_tx_legacy(transaction=transaction, signers=signers)

    def _sign_ixs(self, ix_container: JSONLike) -> JSONLike:
        """Create a signed transaction from instructions."""
        tx = SolanaTransaction(
            fee_payer=self.entity.pubkey(),
            recent_blockhash=Hash.from_string(ix_container["recent_blockhash"]),
            instructions=[
                Instruction.from_json(json.dumps(ix)) for ix in ix_container["ixs"]
            ],
        )
        tx.sign(self.entity)
        return json.loads(tx.to_solders().to_json())

    def _sign_tx_legacy(
        self, transaction: JSONLike, signers: Optional[list] = None
    ) -> JSONLike:
        """
        Sign a transaction in bytes string form.

        :param transaction: the transaction to be signed
        :param signers: list of signers
        :return: signed transaction
        """

        signers = signers or []
        signers = [Keypair.from_base58_string(signer.private_key) for signer in signers]
        signers.append(self.entity)

        recent_hash = Hash.from_string(transaction["recentBlockhash"])
        if isinstance(transaction["message"], list):
            transaction["message"][1]["recentBlockhash"] = recent_hash
            return json.loads(
                VersionedTransaction(
                    message=MessageV0.from_json(json.dumps(transaction["message"][1])),
                    keypairs=signers,
                ).to_json()
            )

        stxn = Transaction.from_json(json.dumps(transaction))
        stxn.sign(keypairs=signers, recent_blockhash=recent_hash)
        return json.loads(stxn.to_json())

    @classmethod
    def generate_private_key(
        cls, extra_entropy: Union[str, bytes, int] = ""
    ) -> Keypair:
        """
        Generate a key pair for Solana network.

        :param extra_entropy: add extra randomness to whatever randomness your OS can provide
        :return: keypair object
        """
        if extra_entropy:
            raise ValueError("extra_entropy is not supported at the moment")
        account = Keypair()  # pylint: disable=no-value-for-parameter
        return account

    def encrypt(self, password: str) -> str:
        """
        Encrypt the private key and return in json.

        :param password: the password to decrypt.
        :return: json string containing encrypted private key.
        """
        try:
            pw = str.encode(password)
            hash_object = hashlib.sha256(pw)
            hex_dig = hash_object.digest()
            base64_bytes = base64.b64encode(hex_dig)
            fernet = Fernet(base64_bytes)
            enc_mac = fernet.encrypt(self.private_key.encode())
        except Exception as e:
            raise Exception("Encryption failed") from e

        return json.dumps(enc_mac.decode())

    @classmethod
    def decrypt(cls, keyfile_json: str, password: str) -> str:
        """
        Decrypt the private key and return in raw form.

        :param keyfile_json: json str containing encrypted private key.
        :param password: the password to decrypt.
        :return: the raw private key.
        """
        try:
            keyfile = json.loads(keyfile_json)
            keyfile_bytes = keyfile.encode()
            pw = str.encode(password)
            hash_object = hashlib.sha256(pw)
            hex_dig = hash_object.digest()
            base64_bytes = base64.b64encode(hex_dig)
            fernet = Fernet(base64_bytes)

            dec_mac = fernet.decrypt(keyfile_bytes).decode()
        except ValueError as e:
            raise DecryptError() from e
        return dec_mac



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the transaction helper for the solana module."""
import json

from solana.transaction import Transaction
from solders.transaction import Transaction as SoldersTransaction
from solders.transaction import VersionedTransaction as BaseSoldersVersionedTransaction


SoldersVersionedTransaction = BaseSoldersVersionedTransaction


class SolanaTransaction(Transaction):
    """Class to represent a solana ledger transaction."""

    @classmethod
    def from_json(cls, json_data: dict) -> "SolanaTransaction":
        """Convert from a json."""
        string_value = json.dumps(json_data)
        solders_tx = SoldersTransaction.from_json(string_value)
        return cls.from_solders(solders_tx)

    def to_json(self) -> dict:
        """Convert to json."""
        return json.loads(self._solders.to_json())



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Conftest module for Pytest."""
import inspect
import logging
import os
import platform
import shutil
import tempfile
from functools import wraps
from pathlib import Path
from typing import Callable

import pytest
from aea_ledger_solana import SolanaCrypto

from aea.configurations.constants import PRIVATE_KEY_PATH_SCHEMA


CUR_PATH = os.path.dirname(inspect.getfile(inspect.currentframe()))  # type: ignore
ROOT_DIR = os.path.join(CUR_PATH, "..")
MAX_FLAKY_RERUNS = 3
AIRDROP_AMOUNT = 1

SOLANA = SolanaCrypto.identifier

SOLANA_PRIVATE_KEY_FILE = PRIVATE_KEY_PATH_SCHEMA.format(SOLANA)

SOLANA_PRIVATE_KEY_PATH = os.path.join(
    ROOT_DIR, "tests", "data", SOLANA_PRIVATE_KEY_FILE
)

SOLANA_PRIVATE_KEY_FILE_1 = os.path.join(
    ROOT_DIR, "tests", "data", SOLANA_PRIVATE_KEY_FILE[:-4] + "_1" + ".txt"
)


SOLANA_DEFAULT_ADDRESS = "http://127.0.0.1:8899"
SOLANA_DEFAULT_CHAIN_ID = 420
SOLANA_DEFAULT_CURRENCY_DENOM = "lamports"
SOLANA_TESTNET_CONFIG = {"address": SOLANA_DEFAULT_ADDRESS}


logger = logging.getLogger(__name__)

PROGRAM_KEYPAIR_PATH = Path(ROOT_DIR, "tests", "data", "solana_private_key_program.txt")
PAYER_KEYPAIR_PATH = Path(ROOT_DIR, "tests", "data", "solana_private_key1.txt")
PLAYER1_KEYPAIR_PATH = Path(ROOT_DIR, "tests", "data", "solana_private_key1.txt")
PLAYER2_KEYPAIR_PATH = Path(ROOT_DIR, "tests", "data", "solana_private_key2.txt")


def action_for_platform(platform_name: str, skip: bool = True) -> Callable:
    """
    Decorate a pytest class or method to skip on certain platform.

    :param platform_name: check `platform.system()` for available platforms.
    :param skip: if True, the test will be skipped; if False, the test will be run ONLY on the chosen platform.
    :return: decorated object
    """

    # for docstyle.
    def decorator(pytest_func):
        """
        For the sake of clarity, assume the chosen platform for the action is "Windows".

        If the following condition is true:
          - the current system is not Windows (is_different) AND we want to skip it (skip)
         OR
          - the current system is Windows (not is_different) AND we want to run only on it (not skip)
        we run the test, else we skip the test.

        logically, the condition is a boolean equivalence
        between the variables "is_different" and "skip"
        Hence, the condition becomes:

        :param pytest_func: the pytest function to wrap
        :return: the wrapped function
        """
        is_different = platform.system() != platform_name
        if is_different is skip:
            return pytest_func

        def action(*args, **kwargs):
            if skip:
                pytest.skip(
                    f"Skipping the test since it doesn't work on {platform_name}."
                )
            else:
                pytest.skip(
                    f"Skipping the test since it works only on {platform_name}."
                )

        if isinstance(pytest_func, type):
            return type(
                pytest_func.__name__,
                (pytest_func,),
                {
                    "setup_class": action,
                    "setup": action,
                    "setUp": action,
                    "_skipped": True,
                },
            )

        @wraps(pytest_func)
        def wrapper(*args, **kwargs):  # type: ignore
            action(*args, **kwargs)

        return wrapper

    return decorator


@pytest.fixture(scope="session")
def solana_private_key_file():
    """Pytest fixture to create a temporary Solana private key file."""
    crypto = SolanaCrypto()
    temp_dir = Path(tempfile.mkdtemp())
    try:
        temp_file = temp_dir / "private_key.txt"
        temp_file.write_text(crypto.private_key)
        yield str(temp_file)

    finally:
        shutil.rmtree(temp_dir)


@pytest.fixture(scope="session")
def solana_testnet_config(ganache_addr, ganache_port):
    """Get Solana ledger api configurations using Ganache."""
    new_uri = f"{ganache_addr}:{ganache_port}"
    new_config = {
        "address": new_uri,
        "chain_id": SOLANA_DEFAULT_CHAIN_ID,
        "denom": SOLANA_DEFAULT_CURRENCY_DENOM,
    }
    return new_config



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023-2024 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the helpers for the solana ledger."""
import hashlib
import json
import time
from pathlib import Path
from typing import Any, Dict, Optional, Tuple, cast

from aea_ledger_solana.transaction import SolanaTransaction
from aea_ledger_solana.utils import pako_inflate
from anchorpy.coder.accounts import ACCOUNT_DISCRIMINATOR_SIZE
from anchorpy.idl import _decode_idl_account
from solana.blockhash import BlockhashCache  # type: ignore
from solana.rpc.api import Client  # type: ignore
from solders.pubkey import Pubkey as PublicKey

from aea.common import Address, JSONLike
from aea.crypto.base import Crypto, Helper
from aea.helpers.base import try_decorator


class SolanaHelper(Helper):
    """Helper class usable as Mixin for SolanaApi or as standalone class."""

    BlockhashCache: BlockhashCache  # defined in SolanaAPi.__init__
    _api: Client  # defined in SolanaAPi.__init__

    @classmethod
    def load_contract_interface(
        cls,
        idl_file_path: Optional[Path] = None,
        program_keypair: Optional[Crypto] = None,
        program_address: Optional[str] = None,
        rpc_api: Optional[str] = None,
        bytecode_path: Optional[Path] = None,
    ) -> Dict[str, Any]:  # type: ignore
        """
        Load contract interface.

        :param idl_file_path: the file path to the IDL
        :param program_keypair: the program keypair
        :param program_address: the program address
        :param rpc_api: the rpc api
        :param bytecode_path: the file path to the bytecode

        :return: the interface
        """
        if bytecode_path is not None:
            in_file = open(bytecode_path, "rb")
            bytecode = in_file.read()
        else:
            bytecode = None

        if (
            program_keypair is not None or program_address is not None
        ) and rpc_api is not None:
            try:
                pid = program_address or program_keypair.address  # type: ignore # mypy doesnt recognize `if` condition above

                base = PublicKey.find_program_address([], PublicKey(pid))[0]
                idl_address = PublicKey.create_with_seed(
                    base, "anchor:idl", PublicKey(pid)
                )
                client = Client(endpoint=rpc_api)
                account_info = client.get_account_info(idl_address)

                account_info_val = account_info.value
                idl_account = _decode_idl_account(
                    bytes(account_info_val.data)[ACCOUNT_DISCRIMINATOR_SIZE:]
                )
                inflated_idl = pako_inflate(bytes(idl_account["data"])).decode()
                json_idl = json.loads(inflated_idl)
                return {
                    "idl": json_idl,
                    "bytecode": bytecode,
                    "program_address": program_address,
                    "program_keypair": program_keypair,
                }
            except Exception as e:
                raise Exception("Could not locate IDL") from e

        elif idl_file_path is not None:
            with open(idl_file_path, "r") as interface_file_solana:
                json_idl = json.load(interface_file_solana)

            return {
                "idl": json_idl,
                "bytecode": bytecode,
                "program_address": program_address,
                "program_keypair": program_keypair,
            }
        else:
            raise Exception("Could not locate IDL")

    @staticmethod
    def is_transaction_valid(
        tx: JSONLike,
        seller: Address,
        client: Address,
        tx_nonce: str,
        amount: int,
    ) -> bool:
        """
        Check whether a transaction is valid or not.

        :param tx: the transaction.
        :param seller: the address of the seller.
        :param client: the address of the client.
        :param tx_nonce: the transaction nonce.
        :param amount: the amount we expect to get from the transaction.

        # noqa: DAR202

        :return: True if the random_message is equals to tx['input']
        """
        raise NotImplementedError

    @staticmethod
    def is_transaction_settled(tx_receipt: JSONLike) -> bool:
        """
        Check whether a transaction is settled or not.

        :param tx_receipt: the receipt associated to the transaction.
        :return: True if the transaction has been settled, False o/w.
        """
        is_successful = False
        if tx_receipt is not None:
            is_successful = tx_receipt["meta"]["status"] == {"Ok": None}  # type: ignore
        return is_successful

    @staticmethod
    def get_hash(message: bytes) -> str:
        """
        Get the hash of a message.

        :param message: the message to be hashed.
        :return: the hash of the message as a hex string.
        """
        sha = hashlib.sha256()
        sha.update(message)
        return sha.hexdigest()

    @classmethod
    def recover_message(
        cls, message: bytes, signature: str, is_deprecated_mode: bool = False
    ) -> Tuple[Address, ...]:
        """
        Recover the addresses from the hash.

        **TOBEIMPLEMENTED**

        :param message: the message we expect
        :param signature: the transaction signature
        :param is_deprecated_mode: if the deprecated signing was used

        # noqa: DAR202

        :return: the recovered addresses
        """
        raise NotImplementedError

    @classmethod
    def recover_public_keys_from_message(
        cls, message: bytes, signature: str, is_deprecated_mode: bool = False
    ) -> Tuple[str, ...]:
        """
        Get the public key used to produce the `signature` of the `message`

        **TOBEIMPLEMENTED**

        :param message: raw bytes used to produce signature
        :param signature: signature of the message
        :param is_deprecated_mode: if the deprecated signing was used

        # noqa: DAR202

        :return: the recovered public keys
        """

        raise NotImplementedError

    @try_decorator("Unable to get nonce: {}", logger_method="warning")
    def _generate_tx_nonce(self) -> str:
        """
        Fetch a latest blockhash to distinguish transactions with the same terms.

        :return: return the blockhash as a nonce.
        """

        try:
            blockhash = self._api.BlockhashCache.get()
            return blockhash
        except Exception:  # pylint: disable=broad-except
            result = self._api.get_latest_blockhash()
            blockhash_json = result.value.to_json()
            blockhash = json.loads(blockhash_json)
            self.BlockhashCache.set(
                blockhash=blockhash["blockhash"], slot=result.context.slot
            )
            return blockhash["blockhash"]

    @staticmethod
    def generate_tx_nonce(seller: Address, client: Address) -> str:
        """
        Generate a unique hash to distinguish transactions with the same terms.

        :param seller: the address of the seller.
        :param client: the address of the client.
        :return: return the hash in hex.
        """
        time_stamp = int(time.time())
        aggregate_hash = hashlib.sha256(
            b"".join([seller.encode(), client.encode(), time_stamp.to_bytes(32, "big")])
        )
        return aggregate_hash.hexdigest()

    @staticmethod
    def to_transaction_format(tx: dict) -> Any:
        """Check whether a transaction is valid or not."""
        jsonTx = json.dumps(tx)
        stxn = SolanaTransaction.from_json(jsonTx)  # mypy: ignore
        return SolanaTransaction.from_solders(stxn)

    @staticmethod
    def to_dict_format(tx) -> JSONLike:
        """
        Check whether a transaction is valid or not.

        :param tx: the transaction.
        :return: True if the random_message is equals to tx['input']
        """

        return json.loads(tx._solders.to_json())  # pylint: disable=protected-access

    @staticmethod
    def get_contract_address(tx_receipt: JSONLike) -> Optional[str]:
        """
        Retrieve the `contract_addresses` from a transaction receipt.

        **Solana can have many contract addresses in one tx**

        :param tx_receipt: the receipt of the transaction.
        :return: the contract address, if present
        """
        contract_addresses = []
        keys = tx_receipt["transaction"]["message"]["accountKeys"]  # type: ignore
        for ix in tx_receipt["transaction"]["message"]["instructions"]:  # type: ignore
            program_index = ix["programIdIndex"]  # type: ignore
            contract_addresses.append(cast(str, keys[program_index]))  # type: ignore
        return contract_addresses[0] if contract_addresses else None

    @classmethod
    def get_address_from_public_key(cls, public_key: str) -> str:
        """
        Get the address from the public key.

        :param public_key: the public key
        :return: str
        """
        return str(PublicKey(bytes.fromhex(public_key)))

    @classmethod
    def is_valid_address(cls, address: str) -> bool:
        """
        Check if the address is valid.

        :param address: the address to validate
        :return: whether the address is valid
        """
        try:
            PublicKey.from_string(address)
            return True
        except Exception:  # pylint: disable=broad-except
            return False



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the tests of the solana module."""

import pytest
from aea_ledger_solana import SolanaApi, SolanaCrypto, SolanaFaucetApi
from solana.transaction import Transaction
from solders import system_program as sp
from solders.keypair import Keypair
from solders.pubkey import Pubkey
from solders.system_program import CreateAccountWithSeedParams
from solders.system_program import ID as SYS_PROGRAM_ID
from solders.system_program import TransferParams, transfer


@pytest.fixture
def test_client():
    """Create a client for testing."""
    return SolanaApi()


#


@pytest.mark.integration
def test_send_transaction_and_get_balance(
    test_client,
):
    """Test sending a transaction to localnet."""

    sender = SolanaCrypto()
    receiver = SolanaCrypto()
    # we need to fund the new sender
    faucet = SolanaFaucetApi()
    faucet.generate_wealth_if_needed(test_client, sender.address)
    # we need to interact with the SPL token program so that we can transfer lamports
    transfer_ix = transfer(
        TransferParams(
            from_pubkey=sender.entity.pubkey(),
            to_pubkey=receiver.entity.pubkey(),
            lamports=1_000_000,
        )
    )
    txn = Transaction().add(transfer_ix)
    resp = test_client.api.send_transaction(
        txn, Keypair.from_base58_string(sender.private_key)
    )
    tx_digest = str(resp.value)
    test_client.wait_get_receipt(tx_digest)
    # check balance
    balance = test_client.api.get_balance(receiver.entity.pubkey()).value
    assert balance == 1_000_000


#


@pytest.mark.integration
def test_create_program_account(
    test_client,
):
    """Test sending a transaction to localnet."""
    # Create transfer tx to transfer 1m lamports from sender to receiver

    sender = SolanaCrypto()
    receiver = SolanaCrypto()
    # we need to fund the new sender
    faucet = SolanaFaucetApi()
    faucet.generate_wealth_if_needed(test_client, sender.address)
    # we need to interact with the SPL token program so that we can transfer lamports
    transfer_ix = transfer(
        TransferParams(
            from_pubkey=sender.entity.pubkey(),
            to_pubkey=receiver.entity.pubkey(),
            lamports=1_000_000,
        )
    )
    txn = Transaction().add(transfer_ix)
    resp = test_client.api.send_transaction(
        txn, Keypair.from_base58_string(sender.private_key)
    )
    tx_digest = str(resp.value)
    test_client.wait_get_receipt(tx_digest)
    # check balance
    balance = test_client.api.get_balance(receiver.entity.pubkey()).value
    assert balance == 1_000_000


@pytest.mark.unit
def test_create_account():
    """Test the create_account function."""
    params = sp.CreateAccountParams(
        from_pubkey=Keypair().pubkey(),
        to_pubkey=Keypair().pubkey(),
        lamports=123,
        space=1,
        owner=Pubkey.default(),
    )
    assert sp.decode_create_account(sp.create_account(params)) == params


@pytest.mark.integration
def test_submit_create_account(test_client):
    """Test the create_account function."""

    sender = SolanaCrypto()
    sender_kp = Keypair.from_base58_string(sender.private_key)
    faucet = SolanaFaucetApi()
    solana_api = SolanaApi()
    seed = "12123123"
    acc = Pubkey.create_with_seed(
        sender.entity.pubkey(),
        seed,
        SYS_PROGRAM_ID,
    )
    amount = 1000023
    params = CreateAccountWithSeedParams(
        from_pubkey=Pubkey.from_string(sender.address),
        to_pubkey=acc,
        base=Pubkey.from_string(sender.address),
        seed=seed,
        lamports=amount,
        space=0,
        owner=SYS_PROGRAM_ID,
    )
    ix_create_pda = sp.create_account_with_seed(params)
    txn = Transaction(fee_payer=Pubkey.from_string(sender.address)).add(ix_create_pda)
    faucet.generate_wealth_if_needed(solana_api, sender.address)
    resp = test_client.api.send_transaction(txn, sender_kp)
    tx_digest = str(resp.value)
    result = test_client.wait_get_receipt(tx_digest)
    assert result[1]



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""Solana account implementation."""
from dataclasses import dataclass

from solders.instruction import AccountMeta
from solders.pubkey import Pubkey as PublicKey


@dataclass
class AccountMeta:
    """Account metadata dataclass."""

    pubkey: PublicKey
    """An account's public key."""
    is_signer: bool
    """True if an instruction requires a transaction signature matching `pubkey`"""
    is_writable: bool
    """True if the `pubkey` can be loaded as a read-write account."""

    @classmethod
    def from_solders(cls, meta: AccountMeta):
        """Convert from a `solders` AccountMeta."""
        return cls(
            pubkey=PublicKey.from_bytes(bytes(meta.pubkey)),
            is_signer=meta.is_signer,
            is_writable=meta.is_writable,
        )

    def to_solders(self) -> AccountMeta:
        """Convert to a `solders` AccountMeta."""
        return AccountMeta(
            pubkey=self.pubkey, is_signer=self.is_signer, is_writable=self.is_writable
        )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the Faucet implementation for the solana ledger."""
import json
import time
from typing import Optional, Union

from aea_ledger_solana.constants import (
    DEFAULT_ADDRESS,
    LAMPORTS_PER_SOL,
    TESTNET_NAME,
    _SOLANA,
)
from aea_ledger_solana.utils import default_logger
from solana.rpc.api import Client  # type: ignore
from solders.pubkey import Pubkey as PublicKey

from aea.common import Address
from aea.crypto.base import FaucetApi
from aea.helpers.base import try_decorator


class SolanaFaucetApi(FaucetApi):
    """Solana testnet faucet API."""

    identifier = _SOLANA
    testnet_name = TESTNET_NAME
    DEFAULT_AMOUNT = 500000000

    def get_wealth(self, address: Address, url: Optional[str] = None) -> None:
        """
        Get wealth from the faucet for the provided address.

        :param address: the address.
        :param url: the url

        """
        amount = self.DEFAULT_AMOUNT
        self.try_get_wealth(address, amount, url)

    @staticmethod
    @try_decorator(
        "An error occurred while attempting to generate wealth:\n{}",
        logger_method="error",
    )
    def try_get_wealth(
        address: Address, amount: Optional[int] = None, url: Optional[str] = None
    ) -> Optional[str]:
        """
        Get wealth from the faucet for the provided address.

        :param address: the address.
        :param amount: optional int
        :param url: the url

        :return: optional string
        """
        if url is None:
            url = DEFAULT_ADDRESS

        if amount is None:
            amount = int(LAMPORTS_PER_SOL * 0.5)
        else:
            amount = LAMPORTS_PER_SOL * amount

        solana_client = Client(url, commitment="confirmed")
        resp = solana_client.request_airdrop(PublicKey.from_string(address), amount)

        response = json.loads(resp.to_json())
        if "message" in response:
            default_logger.error("Response: {}".format(response["message"]))
            raise Exception(response.get("message"))  # pylint
        if response["result"] is None:
            default_logger.error("Response: {}".format("airdrop failed"))
        elif "error" in response:  # pragma: no cover
            default_logger.error("Response: {}".format("airdrop failed"))
        elif "result" in response:  # pragma: nocover
            default_logger.warning(
                "Response: {}\nMessage: {}".format("success", response["result"])
            )
            return response["result"]
        raise Exception("airdrop failed")

    @staticmethod
    def generate_wealth_if_needed(
        api,
        address,
        min_amount=None,
    ) -> Union[str, None]:
        """Check the balance prior to generating wealth."""
        balance = api.get_balance(address)

        min_balance = min_amount if min_amount is not None else 1000000000
        if balance >= min_balance:
            return "not required"
        faucet = SolanaFaucetApi()
        cnt = 0
        transaction_digest = None
        while transaction_digest is None and cnt < 10:
            transaction_digest = faucet.try_get_wealth(address)
            cnt += 1
            time.sleep(5)

        if transaction_digest is None:
            return "failed"
        _, is_settled = api.wait_get_receipt(transaction_digest)
        if is_settled is True:
            return "success"
        return "failed"



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the tests of the solana module."""
import logging

import pytest
from aea_ledger_solana import SolanaApi, SolanaCrypto

from tests.conftest import MAX_FLAKY_RERUNS


def test_creation(solana_private_key_file):
    """Test the creation of the crypto_objects."""
    assert SolanaCrypto(), "Managed to initialise the solana_keypair"
    assert SolanaCrypto(solana_private_key_file), "Managed to load the sol private key"


def test_derive_address():
    """Test the get_address_from_public_key method"""
    account = SolanaCrypto()
    address = SolanaApi.get_address_from_public_key(account.public_key)
    assert account.address == address, "Address derivation incorrect"


@pytest.mark.flaky(reruns=MAX_FLAKY_RERUNS)
@pytest.mark.integration
@pytest.mark.ledger
def test_encrypt_decrypt_privatekey(caplog, solana_private_key_file):
    """Test the balance is zero for a new account."""
    with caplog.at_level(logging.DEBUG, logger="aea.crypto.solana._default_logger"):
        sc = SolanaCrypto(private_key_path=solana_private_key_file)
        privKey = sc.private_key

        encrypted = sc.encrypt("test123456788")

        decrypted = sc.decrypt(encrypted, "test123456788")
        assert privKey == decrypted, "Private keys match"


def test_can_sign_message():
    """Test the signing of a message."""
    account = SolanaCrypto()
    message = b"hello"
    signature = account.sign_message(message)
    assert signature, "Managed to sign the message"


def test_is_address_string():
    """Test the is_address string method."""
    account = SolanaCrypto()
    assert isinstance(account.address, str)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""Constants for the solana ledger."""
_VERSION = "1.24.17"
_SOLANA = "solana"
TESTNET_NAME = "n/a"
DEFAULT_ADDRESS = "https://api.devnet.solana.com"
DEFAULT_CHAIN_ID = 101
DEFAULT_CURRENCY_DENOM = "lamports"
RENT_EXEMPT_AMOUNT = 1000000

LAMPORTS_PER_SOL = 1000000000
_IDL = "idl"
_BYTECODE = "bytecode"



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the tests of the solana module."""

import logging
import platform
import time
from pathlib import Path
from typing import Optional, Tuple, Union

import pytest
from solders.signature import Signature


if platform.system() != "Linux":
    pytest.skip("Runs only on linux", allow_module_level=True)

from aea_ledger_solana import LAMPORTS_PER_SOL, SolanaApi, SolanaCrypto, SolanaFaucetApi
from nacl.signing import VerifyKey
from solders.system_program import ID as SYS_PROGRAM_ID

from aea.common import JSONLike

from tests.conftest import AIRDROP_AMOUNT, MAX_FLAKY_RERUNS, ROOT_DIR


# testing keys #
program_keypair_path = Path(ROOT_DIR, "tests", "data", "solana_private_key_program.txt")
payer_keypair_path = Path(ROOT_DIR, "tests", "data", "solana_private_key1.txt")
player1_keypair_path = Path(ROOT_DIR, "tests", "data", "solana_private_key1.txt")
player2_keypair_path = Path(ROOT_DIR, "tests", "data", "solana_private_key2.txt")

# helper functions #
EXAMPLE_PUBLIC_ADDRESS = "A8eco8kzjUk5CdefzuM1nZMUDmuZ7Eg4UfAstfCqFNTi"


def retry_airdrop_if_result_none(faucet, address, amount=None):
    """Fund address with faucet."""
    cnt = 0
    tx = None
    while tx is None and cnt < 10:
        tx = faucet.get_wealth(address, amount)
        cnt += 1
        time.sleep(2)
    return tx


def _generate_wealth_if_needed(
    api, address, amount=None, min_amount=None
) -> Union[str, None]:
    balance = api.get_balance(address)

    min_balance = min_amount if min_amount is not None else 1000000000
    if balance >= min_balance:
        return "not required"
    else:
        faucet = SolanaFaucetApi()
        cnt = 0
        transaction_digest = None
        while transaction_digest is None and cnt < 10:
            transaction_digest = faucet.try_get_wealth(address)
            cnt += 1
            time.sleep(10)

        if transaction_digest is None:
            return "failed"
        else:
            _, is_settled = _wait_get_receipt(api, transaction_digest)
            if is_settled is True:
                return "success"
            else:
                return "failed"


def _wait_get_receipt(
    solana_api: SolanaApi, transaction_digest: str
) -> Tuple[Optional[JSONLike], bool]:
    transaction_receipt = None
    not_settled = True
    elapsed_time = 0
    time_to_wait = 40
    sleep_time = 0.25
    while not_settled and elapsed_time < time_to_wait:
        elapsed_time += sleep_time
        time.sleep(sleep_time)
        transaction_receipt = solana_api.get_transaction_receipt(transaction_digest)
        if transaction_receipt is None:
            continue
        is_settled = solana_api.is_transaction_settled(transaction_receipt)
        not_settled = not is_settled

    return transaction_receipt, not not_settled


def _construct_and_settle_tx(
    solana_api: SolanaApi,
    account1: SolanaCrypto,
    account2: SolanaCrypto,
    tx_params: dict,
) -> Tuple[str, JSONLike, bool]:
    """Construct and settle a transaction."""
    transfer_transaction = solana_api.get_transfer_transaction(
        **tx_params, tx_fee=0, tx_nonce=""
    )
    # add nonce
    transfer_transaction = solana_api.add_nonce(transfer_transaction)

    signed_transaction = account1.sign_transaction(transfer_transaction)

    transaction_digest = solana_api.send_signed_transaction(signed_transaction)
    assert transaction_digest is not None, "Failed to submit transfer transaction!"

    transaction_receipt, is_settled = _wait_get_receipt(solana_api, transaction_digest)

    assert transaction_receipt is not None, "Failed to retrieve transaction receipt."

    return transaction_digest, transaction_receipt, is_settled


# tests #


@pytest.mark.flaky(reruns=MAX_FLAKY_RERUNS)
@pytest.mark.integration
@pytest.mark.ledger
def test_get_wealth(caplog, solana_private_key_file):
    """Test the balance is zero for a new account."""
    with caplog.at_level(logging.DEBUG, logger="aea.crypto.solana._default_logger"):
        solana_api = SolanaApi()
        sc = SolanaCrypto(solana_private_key_file)
        resp = _generate_wealth_if_needed(solana_api, sc.address, AIRDROP_AMOUNT)
        assert resp != "failed", "Failed to generate wealth"


@pytest.mark.flaky(reruns=MAX_FLAKY_RERUNS)
@pytest.mark.integration
@pytest.mark.ledger
def test_get_state(solana_private_key_file):
    """Test the get_address_from_public_key method"""

    solana_api = SolanaApi()
    account_state = solana_api.get_state("11111111111111111111111111111111")

    assert hasattr(account_state, "lamports"), "State not in correct format"


def test_creation(solana_private_key_file):
    """Test the creation of the crypto_objects."""
    assert SolanaCrypto(), "Managed to initialise the solana_keypair"
    assert SolanaCrypto(solana_private_key_file), "Managed to load the sol private key"


def test_derive_address():
    """Test the get_address_from_public_key method"""
    account = SolanaCrypto()
    address = SolanaApi.get_address_from_public_key(account.public_key)
    assert account.address == address, "Address derivation incorrect"


def test_get_hash():
    """Test the get hash functionality."""
    expected_hash = "2cf24dba5fb0a30e26e83b2ac5b9e29e1b161e5c1fa7425e73043362938b9824"
    hash_ = SolanaApi.get_hash(message=b"hello")
    assert expected_hash == hash_


def test_is_address_valid():
    """Test the get hash functionality."""
    wallet = SolanaCrypto()
    sa = SolanaApi()
    assert sa.is_valid_address(wallet.address)

    assert not sa.is_valid_address("123IamNotReal")


def test_sign_message():
    """Test message sign functionality."""
    wallet = SolanaCrypto()
    wallet2 = SolanaCrypto()
    msg = bytes("hello", "utf8")
    msg2 = bytes("hellooo", "utf8")
    sig = wallet.sign_message(msg)

    signature = Signature.from_string(sig)

    try:
        verification_key = VerifyKey(bytes(wallet.entity.pubkey()))
        verification_key.verify(smessage=msg2, signature=bytes(signature))
    except Exception as e:
        assert e.args[0] == "Signature was forged or corrupt"

    try:
        verification_key = VerifyKey(bytes(wallet2.entity.pubkey()))
        verification_key.verify(smessage=msg, signature=bytes(signature))
    except Exception as e:
        assert e.args[0] == "Signature was forged or corrupt"

    verification_key = VerifyKey(bytes(wallet.entity.pubkey()))
    result = verification_key.verify(smessage=msg, signature=bytes(signature))

    assert result == msg, "Failed to sign message"


@pytest.mark.flaky(reruns=MAX_FLAKY_RERUNS)
@pytest.mark.integration
@pytest.mark.ledger
def test_load_contract_interface_from_program_id():
    """Test that you can load contract interface from onchain idl store."""
    solana_api = SolanaApi()
    idl_path = Path(
        ROOT_DIR, "tests", "data", "tic-tac-toe", "target", "idl", "tic_tac_toe.json"
    )
    contract_interface = solana_api.load_contract_interface(idl_file_path=idl_path)

    assert "name" in contract_interface["idl"], "idl has a name"


@pytest.mark.flaky(reruns=MAX_FLAKY_RERUNS)
@pytest.mark.integration
@pytest.mark.ledger
def test_unfunded_transfer_transaction(solana_private_key_file):
    """Test the construction, signing and submitting of a transfer transaction."""
    account1 = SolanaCrypto(payer_keypair_path)
    account2 = SolanaCrypto()
    solana_api = SolanaApi()

    SolanaFaucetApi().get_wealth(account1.address)
    assert solana_api.get_balance(account1.address) >= SolanaFaucetApi.DEFAULT_AMOUNT

    AMOUNT = 1232323
    balance1 = solana_api.get_balance(account1.address)
    balance2 = solana_api.get_balance(account2.address)

    assert balance1 >= AMOUNT
    assert balance2 == 0

    tx_params = {
        "sender_address": account1.address,
        "destination_address": account2.address,
        "amount": AMOUNT,
    }

    transaction_digest, transaction_receipt, is_settled = _construct_and_settle_tx(
        solana_api,
        account1,
        account2,
        tx_params,
    )
    assert is_settled, "Failed to verify tx!"

    tx = solana_api.get_transaction(transaction_digest)

    assert tx["blockTime"] == transaction_receipt["blockTime"], "Should be same"

    balance3 = solana_api.get_balance(account2.address)

    assert AMOUNT == balance3, "Should be the same balance"


@pytest.mark.flaky(reruns=MAX_FLAKY_RERUNS, reruns_delay=5)
@pytest.mark.integration
@pytest.mark.ledger
def test_funded_transfer_transaction(solana_private_key_file):
    """Test the construction, signing and submitting of a transfer transaction."""
    account1 = SolanaCrypto(payer_keypair_path)
    account2 = SolanaCrypto(player2_keypair_path)

    solana_api = SolanaApi()
    solana_faucet_api = SolanaFaucetApi()
    solana_faucet_api.get_wealth(account1.address)
    solana_faucet_api.get_wealth(account2.address)

    balance1 = solana_api.get_balance(account1.address)
    balance2 = solana_api.get_balance(account2.address)
    assert balance1 >= solana_faucet_api.DEFAULT_AMOUNT
    assert balance2 >= solana_faucet_api.DEFAULT_AMOUNT

    AMOUNT = 2222

    tx_params = {
        "sender_address": account1.address,
        "destination_address": account2.address,
        "amount": AMOUNT,
    }

    transaction_digest, transaction_receipt, is_settled = _construct_and_settle_tx(
        solana_api,
        account1,
        account2,
        tx_params,
    )
    assert is_settled, "Failed to verify tx!"

    tx = solana_api.get_transaction(transaction_digest)

    assert tx["blockTime"] == transaction_receipt["blockTime"], "Should be same"
    assert balance2 + AMOUNT == solana_api.get_balance(account2.address)


@pytest.mark.flaky(reruns=MAX_FLAKY_RERUNS, reruns_delay=5)
@pytest.mark.integration
@pytest.mark.ledger
def test_get_sol_balance(caplog, solana_private_key_file):
    """Test the balance is zero for a new account."""
    with caplog.at_level(logging.DEBUG, logger="aea.crypto.solana._default_logger"):
        sc = SolanaCrypto(payer_keypair_path)
        sa = SolanaApi()

        balance = sa.get_balance(sc.address)
        assert isinstance(balance, int)


@pytest.mark.flaky(reruns=MAX_FLAKY_RERUNS)
@pytest.mark.integration
@pytest.mark.ledger
def test_get_tx(caplog, solana_private_key_file):
    """Test get tx from signature"""
    with caplog.at_level(logging.DEBUG, logger="aea.crypto.solana._default_logger"):
        solana_faucet_api = SolanaFaucetApi()
        sc = SolanaCrypto(private_key_path=solana_private_key_file)
        solana_api = SolanaApi()

        retries = 0
        tx_signature = None
        while retries <= MAX_FLAKY_RERUNS:
            tx_signature = solana_faucet_api.try_get_wealth(sc.address, AIRDROP_AMOUNT)
            if tx_signature is None:
                retries += 1
                time.sleep(2)
            else:
                break

        assert tx_signature is not None
        tx, settled = _wait_get_receipt(solana_api, tx_signature)
        assert settled is True
        contract_address = solana_api.get_contract_address(tx)
        assert contract_address == "11111111111111111111111111111111"


@pytest.mark.flaky(reruns=MAX_FLAKY_RERUNS)
@pytest.mark.integration
@pytest.mark.ledger
def test_encrypt_decrypt_privatekey(caplog, solana_private_key_file):
    """Test the balance is zero for a new account."""
    with caplog.at_level(logging.DEBUG, logger="aea.crypto.solana._default_logger"):
        sc = SolanaCrypto(private_key_path=solana_private_key_file)
        privKey = sc.private_key

        encrypted = sc.encrypt("test123456788")

        decrypted = sc.decrypt(encrypted, "test123456788")
        assert privKey == decrypted, "Private keys match"


def test_load_contract_interface():
    """Test the load_contract_interface method."""
    path = Path(ROOT_DIR, "tests", "data", "dummy_contract", "build", "idl.json")
    result = SolanaApi.load_contract_interface(path)

    assert "name" in result["idl"]


def test_load_contract_instance():
    """Test the load_contract_interface method."""
    idl_path = Path(
        ROOT_DIR,
        "tests",
        "data",
        "spl-token-faucet",
        "target",
        "idl",
        "spl_token_faucet.json",
    )
    bytecode_path = Path(
        ROOT_DIR,
        "tests",
        "data",
        "spl-token-faucet",
        "target",
        "deploy",
        "spl_token_faucet.so",
    )
    sa = SolanaApi()
    result = sa.load_contract_interface(
        idl_file_path=idl_path, bytecode_path=bytecode_path
    )
    pid = "ZETAxsqBRek56DhiGXrn75yj2NHU3aYUnxvHXpkf3aD"
    instance = SolanaApi.get_contract_instance(
        SolanaApi, contract_interface=result, contract_address=pid
    )

    assert hasattr(instance["program"], "coder")


@pytest.mark.integration
@pytest.mark.ledger
@pytest.mark.flaky(reruns=MAX_FLAKY_RERUNS, reruns_delay=15)
def test_get_transaction_transfer_logs(solana_private_key_file):
    """Test SolanaApi.get_transaction_transfer_logs."""
    solana_api = SolanaApi()

    account1 = SolanaCrypto(payer_keypair_path)

    resp = _generate_wealth_if_needed(solana_api, account1.address)
    assert resp != "failed", "Failed to generate wealth"

    account2 = SolanaCrypto()

    balance1 = solana_api.get_balance(account1.address)
    balance2 = solana_api.get_balance(account2.address)
    assert balance1 > 0
    assert balance2 >= 0

    AMOUNT = 1232323
    tx_params = {
        "sender_address": account1.address,
        "destination_address": account2.address,
        "amount": AMOUNT,
        "unfunded_account": True,
    }

    transaction_digest, transaction_receipt, is_settled = _construct_and_settle_tx(
        solana_api,
        account1,
        account2,
        tx_params,
    )
    assert is_settled, "Failed to verify tx!"

    tx = solana_api.get_transaction(transaction_digest)

    assert tx["blockTime"] == transaction_receipt["blockTime"], "Should be same"

    logs = solana_api.get_transaction_transfer_logs(
        contract_instance=None, tx_hash=transaction_digest
    )
    assert "preBalances" in logs
    assert "postBalances" in logs


@pytest.mark.skip(".get_deploy_transaction not implemented!")
@pytest.mark.flaky(reruns=MAX_FLAKY_RERUNS)
@pytest.mark.integration
@pytest.mark.ledger
def test_deploy_program():
    """Test the deploy contract method."""

    idl_path = Path(
        ROOT_DIR, "tests", "data", "tic-tac-toe", "target", "idl", "tic_tac_toe.json"
    )
    bytecode_path = Path(
        ROOT_DIR, "tests", "data", "tic-tac-toe", "target", "deploy", "tic_tac_toe.so"
    )
    program_keypair_path = Path(
        ROOT_DIR, "tests", "data", "solana_private_key_program.txt"
    )
    payer_keypair_path = Path(ROOT_DIR, "tests", "data", "solana_private_key0.txt")

    sa = SolanaApi()

    program = SolanaCrypto(str(program_keypair_path))
    payer = SolanaCrypto(str(payer_keypair_path))

    interface = sa.load_contract_interface(
        idl_file_path=idl_path, bytecode_path=bytecode_path, program_keypair=program
    )

    init = False
    if init:
        program.dump(str(program_keypair_path))
        payer.dump(str(payer_keypair_path))

        faucet = SolanaFaucetApi()
        tx = retry_airdrop_if_result_none(faucet, payer.address, 1)
        assert tx is not None, "Generate wealth failed"
        _, is_settled = _wait_get_receipt(sa, tx)
        assert is_settled is True

        balance = sa.get_balance(payer.address)
        assert balance >= 2 * LAMPORTS_PER_SOL
        print("Payer Balance: " + str(balance / LAMPORTS_PER_SOL) + " SOL")

    result = sa.get_deploy_transaction(interface, payer)
    assert result is not None, "Should not be none"


@pytest.mark.flaky(reruns=MAX_FLAKY_RERUNS, reruns_delay=5)
@pytest.mark.integration
@pytest.mark.ledger
def test_contract_method_call():
    """Test the deploy contract method."""

    idl_path = Path(
        ROOT_DIR, "tests", "data", "tic-tac-toe", "target", "idl", "tic_tac_toe.json"
    )
    bytecode_path = Path(
        ROOT_DIR, "tests", "data", "tic-tac-toe", "target", "deploy", "tic_tac_toe.so"
    )
    program_keypair_path = Path(
        ROOT_DIR, "tests", "data", "solana_private_key_program.txt"
    )
    payer_keypair_path = Path(ROOT_DIR, "tests", "data", "solana_private_key0.txt")

    sa = SolanaApi()
    payer = SolanaCrypto(str(payer_keypair_path))
    program_kp = SolanaCrypto(str(program_keypair_path))

    interface = sa.load_contract_interface(
        idl_file_path=idl_path, bytecode_path=bytecode_path, program_keypair=program_kp
    )

    instance = sa.get_contract_instance(
        contract_interface=interface, contract_address=program_kp.address
    )

    player2 = SolanaCrypto("./tests/data/solana_private_key2.txt")

    game = SolanaCrypto()

    resp = _generate_wealth_if_needed(sa, payer.address)
    assert resp != "failed", "Failed to generate wealth"

    resp = _generate_wealth_if_needed(sa, player2.address)
    assert resp != "failed", "Failed to generate wealth"

    # setup game
    program = instance["program"]
    program.provider.wallet = payer.entity

    accounts = {
        "game": game.entity.pubkey(),
        "player_one": payer.entity.pubkey(),
        "system_program": SYS_PROGRAM_ID,
    }

    tx = sa.build_transaction(
        program,
        "setup_game",
        method_args={
            "data": (player2.entity.pubkey(),),
            "accounts": accounts,
        },
        tx_args=None,
    )

    tx = sa.add_nonce(tx)

    signed_transaction = game.sign_transaction(tx, [payer])

    transaction_digest = sa.send_signed_transaction(signed_transaction)
    assert transaction_digest is not None

    _, is_settled = _wait_get_receipt(sa, transaction_digest)
    assert is_settled is True
    state = sa.get_state(game.address)
    decoded_state = program.coder.accounts.decode(state.data)
    player1 = payer
    player2 = player2
    column = 0

    # game loop
    while decoded_state.state.index == 0:
        active_player = player2 if decoded_state.turn % 2 == 0 else player1
        row = 0 if decoded_state.turn % 2 == 0 else 1
        accounts = {
            "game": game.entity.pubkey(),
            "player": active_player.entity.pubkey(),
        }

        tile = program.type["Tile"](row=row, column=column)

        tx1 = sa.build_transaction(
            program,
            "play",
            method_args={
                "data": (tile,),
                "accounts": accounts,
            },
            tx_args=None,
        )

        tx1["message"]["header"]["numReadonlySignedAccounts"] = 0
        tx1 = sa.add_nonce(tx1)

        signed_transaction = active_player.sign_transaction(
            tx1,
        )

        transaction_digest = sa.send_signed_transaction(signed_transaction)
        assert transaction_digest is not None
        _, is_settled = _wait_get_receipt(sa, transaction_digest)
        assert is_settled is True
        state = sa.get_state(game.address)
        decoded_state = program.coder.accounts.decode(state.data)

        if row == 0:
            column += 1

    assert str(decoded_state.state.winner) == player1.address


@pytest.mark.flaky(reruns=MAX_FLAKY_RERUNS)
@pytest.mark.integration
def test_get_create_account_tx():
    """Test whether the create account function necessary for the sending of a transaction to new account works."""
    solana_api = SolanaApi()
    account_1 = SolanaCrypto()
    account_2 = SolanaCrypto()
    tx = solana_api._api.get_create_account_instructions(
        sender_address=account_1.address,
        destination_address=account_2.address,
        lamports=1,
    )
    assert tx is not None, "Create account transaction is None"



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""Class to test the solana transaction."""
from copy import deepcopy

from aea_ledger_solana import SolanaApi, SolanaTransaction


BLANK_TRANSACTION = {
    "signatures": [[0]],
    "message": {
        "header": {
            "numRequiredSignatures": 0,
            "numReadonlySignedAccounts": 0,
            "numReadonlyUnsignedAccounts": 0,
        },
        "accountKeys": [[0]],
        "recentBlockhash": [
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
        ],
        "instructions": [[0]],
    },
}


def test_solana_transaction():
    """Test the solana transaction."""
    transaction = SolanaTransaction()
    assert transaction is not None


def test_from_json():
    """Test the from_json method."""
    # we first create a json transaction

    transaction = SolanaTransaction.from_json(BLANK_TRANSACTION)
    assert transaction is not None


def test_to_json():
    """Test the to_json method."""
    # we first create a json transaction
    transaction = SolanaTransaction()
    assert transaction.to_json() is not None


def test_from_json_matches_to_json():
    """Test the transaction_matches_transaction method."""
    # we first create a json transaction
    transaction = SolanaTransaction()
    json_string = transaction.to_json()
    transaction2 = SolanaTransaction.from_json(json_string)
    assert transaction2.to_json() == json_string


def test_from_json_matches_to_json_signers():
    """Test the transaction_matches_transaction method."""
    # we first create a json transaction
    new_tx = deepcopy(BLANK_TRANSACTION)
    assert new_tx["message"]["header"]["numReadonlySignedAccounts"] == 0
    solana_api = SolanaApi()
    tx_1 = solana_api.add_nonce(new_tx)
    assert tx_1["message"]["header"]["numReadonlySignedAccounts"] == 0



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023-2025 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the Solana API Client implementation for the solana ledger."""

import json
import secrets
from dataclasses import dataclass

import solders.system_program as sp
from solana.rpc.api import Client as ApiClient
from solders.pubkey import Pubkey as PublicKey
from solders.system_program import ID as SYS_PROGRAM_ID
from solders.system_program import TransferParams, transfer
from spl.token.client import Token as SplClient


@dataclass
class SolanaApiClient(ApiClient):
    """Class to interact with the Solana ledger APIs."""

    def __init__(self, *args, **kwargs):
        """Instantiate the client."""
        super().__init__(*args, **kwargs)

    def transfer(self, *args, **kwargs):
        """Transfer tokens."""

    def get_create_account_instructions(
        self,
        sender_address,
        destination_address,
        lamports: int = 100000,
        space: int = 1,
    ):
        """Create a new account."""
        required_balance = SplClient.get_min_balance_rent_for_exempt_for_account(self)

        seed = str(  # Use secure random for cryptographic seed generation
            secrets.randbelow(1000000000)
        )
        acc = PublicKey.create_with_seed(
            PublicKey.from_string(sender_address),
            seed,
            SYS_PROGRAM_ID,
        )
        params = sp.CreateAccountWithSeedParams(
            from_pubkey=PublicKey.from_string(sender_address),
            to_pubkey=acc,
            base=PublicKey.from_string(sender_address),
            seed=seed,
            lamports=max([lamports, required_balance]),
            space=space,
            owner=SYS_PROGRAM_ID,
        )
        ix_create_pda = sp.create_account_with_seed(params)
        ix_create_pda_json = json.loads(ix_create_pda.to_json())
        return ix_create_pda_json

    def get_transfer_tx(self, from_account, to_account, amount):
        """Create a transfer tx."""
        params = TransferParams(
            from_pubkey=PublicKey.from_string(from_account),
            to_pubkey=PublicKey.from_string(to_account),
            lamports=amount,
        )
        return transfer(params)

    def get_account_state(self, account_address: str):
        """Get account state."""
        account_object = self.get_account_info_json_parsed(
            PublicKey.from_string(account_address)
        )
        account_info_val = account_object.value
        # note we dont currently parse this back as json.
        return account_info_val



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the tests of the solana module."""
import logging

import pytest
from aea_ledger_solana import SolanaApi, SolanaCrypto, SolanaFaucetApi

from tests.conftest import AIRDROP_AMOUNT


@pytest.mark.ledger
@pytest.mark.integration
def test_get_wealth_succeeds(caplog, solana_private_key_file):
    """Test the faucet request succeeds."""
    with caplog.at_level(logging.DEBUG, logger="aea.crypto.solana._default_logger"):
        solana_api = SolanaApi()
        solana_faucet = SolanaFaucetApi()
        sc = SolanaCrypto(solana_private_key_file)
        resp = solana_faucet.generate_wealth_if_needed(
            solana_api, sc.address, AIRDROP_AMOUNT
        )
        assert resp != "failed", "Failed to generate wealth"


@pytest.mark.ledger
@pytest.mark.integration
def test_get_wealth_increments_native_balance(caplog, solana_private_key_file):
    """Test the balance increases after faucet request."""
    solana_api = SolanaApi()
    solana_faucet = SolanaFaucetApi()
    sc = SolanaCrypto()
    pre_balance = solana_api.get_balance(sc.address)
    with caplog.at_level(logging.DEBUG, logger="aea.crypto.solana._default_logger"):
        resp = solana_faucet.generate_wealth_if_needed(
            solana_api, sc.address, AIRDROP_AMOUNT
        )
        assert resp != "failed", "Failed to generate wealth"
    assert solana_api.get_balance(sc.address) > pre_balance



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""Tests to ensure the contracts work as expected."""
import platform
from pathlib import Path

import anchorpy
import pytest
from aea_ledger_solana import SolanaApi, SolanaCrypto, SolanaFaucetApi

from tests.conftest import ROOT_DIR


PAYER_KEYPAIR_PATH_0 = Path(ROOT_DIR, "tests", "data", "solana_private_key0.txt")
PAYER_KEYPAIR_PATH_1 = Path(ROOT_DIR, "tests", "data", "solana_private_key1.txt")
PROGRAM_KEYPAIR_PATH = Path(ROOT_DIR, "tests", "data", "solana_private_key_program.txt")


@pytest.fixture
def solana_faucet():
    """Create a solana faucet."""
    sf = SolanaFaucetApi()
    return sf


@pytest.fixture
def payer_1():
    """Create a payer."""
    payer = SolanaCrypto(str(PAYER_KEYPAIR_PATH_0))
    return payer


@pytest.fixture
def payer_2():
    """Create a payer."""
    payer = SolanaCrypto(str(PAYER_KEYPAIR_PATH_1))
    return payer


@pytest.fixture
def solana_api():
    """Create a solana api."""
    sa = SolanaApi()
    return sa


def _get_token_contract(solana_api):
    """Create a contract."""
    idl_path = Path(
        ROOT_DIR,
        "tests",
        "data",
        "spl-token-faucet",
        "target",
        "idl",
        "spl_token_faucet.json",
    )
    bytecode_path = Path(
        ROOT_DIR,
        "tests",
        "data",
        "spl-token-faucet",
        "target",
        "deploy",
        "spl_token_faucet.so",
    )
    program_key_pair = SolanaCrypto(str(PROGRAM_KEYPAIR_PATH))

    interface = solana_api.load_contract_interface(
        idl_file_path=idl_path, bytecode_path=bytecode_path
    )
    instance = solana_api.get_contract_instance(
        contract_interface=interface,
        contract_address="11111111111111111111111111111110",
    )

    return instance, interface, program_key_pair


def _get_tic_tac_contract(solana_api):
    """Create a contract."""
    idl_path = Path(
        ROOT_DIR, "tests", "data", "tic-tac-toe", "target", "idl", "tic_tac_toe.json"
    )
    bytecode_path = Path(
        ROOT_DIR, "tests", "data", "tic-tac-toe", "target", "deploy", "tic_tac_toe.so"
    )
    program_key_pair = SolanaCrypto(str(PROGRAM_KEYPAIR_PATH))

    interface = solana_api.load_contract_interface(
        idl_file_path=idl_path,
        bytecode_path=bytecode_path,
        program_keypair=program_key_pair,
    )
    instance = solana_api.get_contract_instance(
        contract_interface=interface,
        contract_address=program_key_pair.address,
    )

    return instance, interface, program_key_pair


@pytest.mark.skipif(platform.system() == "Windows", reason="Windows not supported")
def test_tic_tac_contract(solana_api):
    """Test the tic tac contract."""
    instance, interface, _ = _get_tic_tac_contract(solana_api)
    assert isinstance(instance["program"], anchorpy.program.core.Program)
    assert isinstance(instance["program"].provider, anchorpy.provider.Provider)
    assert isinstance(interface, dict)


@pytest.fixture
def tic_tac_contract(solana_api):
    """Create a contract."""
    return _get_tic_tac_contract(solana_api)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023-2025 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""Solana module wrapping the public and private key cryptography and ledger api."""
import json
import logging
import time
from typing import Any, Dict, List, Optional, Sequence, Tuple, Union, cast

from aea_ledger_solana.constants import (
    DEFAULT_ADDRESS,
    DEFAULT_CHAIN_ID,
    _SOLANA,
    _VERSION,
)
from aea_ledger_solana.crypto import SolanaCrypto
from aea_ledger_solana.faucet import SolanaFaucetApi  # noqa: F401
from aea_ledger_solana.helper import SolanaHelper
from aea_ledger_solana.solana_api import SolanaApiClient
from aea_ledger_solana.transaction import (
    SolanaTransaction,
    SoldersTransaction,
    SoldersVersionedTransaction,
)
from aea_ledger_solana.transaction_instruction import TransactionInstruction
from anchorpy import Context, Idl, Program  # type: ignore
from solana.blockhash import BlockhashCache
from solana.transaction import Transaction  # type: ignore
from solana.transaction import Transaction as BaseSolanaTransaction
from solders import system_program as ssp  # type: ignore
from solders.hash import Hash
from solders.instruction import AccountMeta, Instruction
from solders.keypair import Keypair
from solders.message import MessageV0
from solders.pubkey import Pubkey  # type: ignore
from solders.pubkey import Pubkey as PublicKey
from solders.signature import Signature  # type: ignore
from solders.system_program import (  # type: ignore; SYS_PROGRAM_ID,
    CreateAccountParams,
    CreateAccountWithSeedParams,
)
from solders.system_program import ID as SYS_PROGRAM_ID  # type: ignore
from solders.system_program import (  # type: ignore; SYS_PROGRAM_ID,
    TransferParams,
    create_account,
    transfer,
)

from aea.common import Address, JSONLike
from aea.crypto.base import LedgerApi
from aea.helpers.base import try_decorator


DEFAULT_MAX_SUPPORTED_TRANSACTION_VERSION = 0
DEFAULT_MAX_RETRIES = 3


class SolanaApi(LedgerApi, SolanaHelper):
    """Class to interact with the Solana Web3 APIs."""

    identifier = _SOLANA

    def __init__(self, **kwargs: Any):
        """
        Initialize the Solana ledger APIs.

        :param kwargs: keyword arguments
        """

        """Type for commitment."""

        self._api = SolanaApiClient(
            endpoint=kwargs.pop("address", DEFAULT_ADDRESS),
        )

        self._chain_id = kwargs.pop("chain_id", DEFAULT_CHAIN_ID)
        self._version = _VERSION
        self.BlockhashCache = BlockhashCache(ttl=10)
        self._get_latest_hash()

    def _get_latest_hash(self):
        """Get the latest block hash."""
        result = self._api.get_latest_blockhash()
        blockhash_json = result.value.to_json()
        blockhash = json.loads(blockhash_json)
        self._hash = blockhash["blockhash"]
        self.BlockhashCache.set(blockhash=self._hash, slot=result.context.slot)

    @property
    def api(self) -> SolanaApiClient:
        """Get the underlying API object."""
        return self._api

    @property
    def latest_hash(self):
        """Get the latest hash."""
        self._get_latest_hash()
        return self._hash

    @property
    def system_program(self) -> Pubkey:
        """System program."""
        return SYS_PROGRAM_ID

    @staticmethod
    def sol_to_lamp(sol: float) -> int:
        """Solana to lamport value."""
        return int(sol * 1000000000)

    @classmethod
    def to_account_meta(
        cls,
        pubkey: Union[Pubkey, str],
        is_signer: bool,
        is_writable: bool,
    ) -> AccountMeta:
        """To account meta."""
        return AccountMeta(
            pubkey=cls.to_pubkey(pubkey),
            is_signer=is_signer,
            is_writable=is_writable,
        )

    @staticmethod
    def to_pubkey(key: Union[SolanaCrypto, Keypair, Pubkey, str]) -> Pubkey:
        """To pubkey."""
        if isinstance(key, Pubkey):
            return key
        if isinstance(key, Keypair):
            return key.pubkey()
        if isinstance(key, SolanaCrypto):
            return key.entity.pubkey()
        try:
            return Pubkey.from_string(key)
        except BaseException:  # noqa: B036
            return Keypair.from_base58_string(key).pubkey()

    @staticmethod
    def to_keypair(key: Union[SolanaCrypto, Keypair, str]) -> Keypair:
        """To keypair object."""
        if isinstance(key, Keypair):
            return key
        if isinstance(key, SolanaCrypto):
            return key.entity
        return Keypair.from_base58_string(key)

    @staticmethod
    def pda(
        seeds: Sequence[bytes],
        program_id: Pubkey,
    ) -> Pubkey:
        """Create TX PDA"""
        pda, _ = Pubkey.find_program_address(
            seeds=seeds,
            program_id=program_id,
        )
        return pda

    @staticmethod
    def create_pda(
        from_address: str,
        new_account_address: str,
        base_address: str,
        seed: str,
        lamports: int,
        space: int,
        program_id: str,
    ):
        """
        Build a create pda transaction.

        :param from_address: the sender public key
        :param new_account_address: the new account public key
        :param base_address: base address
        :param seed: seed
        :param lamports: the amount of lamports to send
        :param space: the space to allocate
        :param program_id: the program id
        :return: the tx, if present
        """
        params = CreateAccountWithSeedParams(
            PublicKey(from_address),
            PublicKey(new_account_address),
            PublicKey(base_address),
            seed,
            lamports,
            space,
            PublicKey(program_id),
        )
        createPDAInstruction = TransactionInstruction.from_solders(
            ssp.create_account_with_seed(params.to_solders())
        )
        txn = Transaction().add(createPDAInstruction)
        tx = txn._solders.to_json()  # pylint: disable=protected-access
        return json.loads(tx)

    def add_nonce(self, tx: dict) -> JSONLike:
        """
        Check whether a transaction is valid or not.

        :param tx: the transaction.
        :return: True if the random_message is equals to tx['input']
        """
        stxn = self.deserialize_tx(tx=tx)
        nonce = self._generate_tx_nonce()
        txn = stxn.to_json()
        txn["recentBlockhash"] = nonce
        return txn

    def wait_get_receipt(
        self,
        transaction_digest: str,
        max_supported_transaction_version: Optional[int] = None,
    ) -> Tuple[Optional[JSONLike], bool]:
        """Wait for the transaction to be settled and return the receipt."""
        transaction_receipt = None
        not_settled = True
        elapsed_time = 0
        time_to_wait = 40
        sleep_time = 0.25
        while not_settled and elapsed_time < time_to_wait:
            elapsed_time += sleep_time
            time.sleep(sleep_time)
            transaction_receipt = self.get_transaction_receipt(
                transaction_digest,
                max_supported_transaction_version=max_supported_transaction_version,
            )
            if transaction_receipt is None:
                continue
            is_settled = self.is_transaction_settled(transaction_receipt)
            not_settled = not is_settled

        return transaction_receipt, not not_settled

    def construct_and_settle_tx(
        self,
        account1: SolanaCrypto,
        account2: SolanaCrypto,
        tx_params: dict,
    ) -> Tuple[str, JSONLike, bool]:
        """Construct and settle a transaction."""
        transfer_transaction = self.get_transfer_transaction(
            **tx_params, tx_fee=0, tx_nonce=""
        )
        # add nonce
        transfer_transaction = self.add_nonce(transfer_transaction)

        signed_transaction = account1.sign_transaction(transfer_transaction)

        transaction_digest = self.send_signed_transaction(signed_transaction)
        if transaction_digest is None:
            raise Exception("Failed to submit transfer transaction!")

        transaction_receipt, is_settled = self.wait_get_receipt(transaction_digest)

        if transaction_receipt is None:
            raise Exception("Failed to settle transfer transaction!")

        return transaction_digest, transaction_receipt, is_settled

    def update_with_gas_estimate(self, transaction: JSONLike) -> JSONLike:
        """
        Attempts to update the transaction with a gas estimate

        **NOT APPLICABLE**

        :param transaction: the transaction
        :return: the updated transaction
        """

        return transaction

    def get_balance(
        self, address: Address, raise_on_try: bool = False
    ) -> Optional[int]:
        """Get the balance of a given account."""
        return self._try_get_balance(address, raise_on_try=raise_on_try)

    @try_decorator("Unable to retrieve balance: {}", logger_method="warning")
    def _try_get_balance(self, address: Address, **_kwargs: Any) -> Optional[int]:
        """Get the balance of a given account."""
        response = self._api.get_balance(
            PublicKey.from_string(address), commitment="processed"
        )  # pylint: disable=no-member
        return response.value

    def get_state(
        self, callable_name: str, *args: Any, raise_on_try: bool = False, **kwargs: Any
    ) -> Optional[JSONLike]:
        """
        Call a specified function on the underlying ledger API.

        This usually takes the form of a web request to be waited synchronously.

        :param callable_name: the name of the API function to be called.
        :param args: the positional arguments for the API function.
        :param raise_on_try: whether the method will raise or log on error
        :param kwargs: the keyword arguments for the API function.
        :return: the ledger API response.
        """
        return self._get_account_state(
            callable_name, *args, raise_on_try=raise_on_try, **kwargs
        )

    @try_decorator("Unable to get state: {}", logger_method="warning")
    def _get_account_state(  # pylint: disable=unused-argument
        self, address: str, *args: Any, **kwargs: Any
    ) -> Optional[JSONLike]:
        """Try to get account state.."""

        if "raise_on_try" in kwargs:
            logging.info(
                f"popping `raise_on_try` from {self.__class__.__name__}.get_state kwargs"  # pylint: disable=protected-access
            )
            kwargs.pop("raise_on_try")

        return self._api.get_account_state(address)

    def get_transfer_transaction(  # pylint: disable=arguments-differ
        self,
        sender_address: Address,
        destination_address: Address,
        amount: int,
        tx_fee: int,
        tx_nonce: str,
        **kwargs: Any,
    ) -> Optional[JSONLike]:
        """
        Submit a transfer transaction to the ledger.

        :param sender_address: the sender address of the payer.
        :param destination_address: the destination address of the payee.
        :param amount: the amount of wealth to be transferred.
        :param tx_fee: the transaction fee.
        :param tx_nonce: verifies the authenticity of the tx
        :param kwargs: the keyword arguments.
        :return: the transfer transaction
        """
        chain_id = kwargs.get("kwargs", None)
        chain_id = chain_id if chain_id is not None else self._chain_id

        state = self.get_state(destination_address)
        # this is if there is no account yet at the destination address

        if state is None:
            # we need to create the account we first get the tx
            # to create the account
            # then we get the tx to transfer the funds
            # then we combine the two txs
            # then we sign the combined tx
            # then we return the combined tx

            create_account_ixn = self._api.get_create_account_instructions(
                sender_address, destination_address
            )

            transfer_ixn = self._api.get_transfer_tx(
                sender_address, destination_address, amount
            )

            txn = Transaction(
                fee_payer=PublicKey.from_string(sender_address),
            )

            ixn_1 = Instruction.from_json(json.dumps(create_account_ixn))
            ixn_2 = Instruction.from_json(transfer_ixn.to_json())

            # in solana we first create the account then we transfer the funds
            txn.add(ixn_1).add(ixn_2)

        else:
            txn = Transaction(fee_payer=PublicKey.from_string(sender_address)).add(
                transfer(
                    TransferParams(
                        from_pubkey=PublicKey.from_string(sender_address),
                        to_pubkey=PublicKey.from_string(destination_address),
                        lamports=amount,
                    )
                )
            )

        tx = txn.to_solders().to_json()  # pylint: disable=protected-access
        return json.loads(tx)

    def send_signed_transaction(
        self, tx_signed: JSONLike, raise_on_try: bool = False
    ) -> Optional[str]:
        """
        Send a signed transaction and wait for confirmation.

        :param tx_signed: the signed transaction
        :param raise_on_try: whether the method will raise or log on error
        :return: tx_digest, if present
        """
        return self._try_send_signed_transaction(tx_signed, raise_on_try=raise_on_try)

    @try_decorator("Unable to send transaction: {}", logger_method="warning")
    def _try_send_signed_transaction(
        self, tx_signed: JSONLike, **_kwargs: Any
    ) -> Optional[str]:
        """
        Try send a signed transaction.

        :param tx_signed: the signed transaction
        :param _kwargs: the keyword arguments. Possible kwargs are:
            `raise_on_try`: bool flag specifying whether the method will raise or log on error (used by `try_decorator`)
        :return: tx_digest, if present
        """
        stxn = self.deserialize_tx(tx=tx_signed)
        txn_resp = self._api.send_raw_transaction(bytes(stxn.serialize()))
        return str(txn_resp.value)

    def send_signed_transactions(
        self,
        signed_transactions: List[JSONLike],
        raise_on_try: bool = False,
        **kwargs: Any,
    ) -> Optional[List[str]]:
        """
        Atomically send multiple of transactions.

        :param signed_transactions: the signed transactions to bundle together and send.
        :param raise_on_try: whether the method will raise or log on error
        :param kwargs: the keyword arguments.
        """
        raise NotImplementedError

    def get_transaction_receipt(
        self,
        tx_digest: str,
        max_supported_transaction_version: Optional[int] = None,
        retries: Optional[int] = None,
        raise_on_try: bool = False,
    ) -> Optional[JSONLike]:
        """
        Get the transaction receipt for a transaction digest.

        :param tx_digest: the digest associated to the transaction.
        :param max_supported_transaction_version: The max transaction version to return in responses.
        :param retries: The max amount of retries for fetching the receipt.
        :param raise_on_try: whether the method will raise or log on error
        :return: the tx receipt, if present
        """
        tx_receipt = self._try_get_transaction_receipt(
            tx_digest,
            max_supported_transaction_version=max_supported_transaction_version,
            retries=retries,
            raise_on_try=raise_on_try,
        )

        return tx_receipt

    @try_decorator(
        "Error when attempting getting tx receipt: {}", logger_method="debug"
    )
    def _try_get_transaction_receipt(
        self,
        tx_digest: str,
        max_supported_transaction_version: Optional[int] = None,
        retries: Optional[int] = None,
        **_kwargs: Any,
    ) -> Optional[JSONLike]:
        """
        Try get the transaction receipt.

        :param tx_digest: the digest associated to the transaction.
        :param max_supported_transaction_version: The max transaction version to return in responses.
        :param retries: The max amount of retries for fetching the receipt.
        :param _kwargs: the keyword arguments. Possible kwargs are:
            `raise_on_try`: bool flag specifying whether the method will raise or log on error (used by `try_decorator`)
        :return: the tx receipt, if present
        """
        retries = retries or DEFAULT_MAX_RETRIES
        while retries > 0:
            receipt = self._api.get_transaction(
                Signature.from_string(tx_digest),
                max_supported_transaction_version=(
                    max_supported_transaction_version
                    or DEFAULT_MAX_SUPPORTED_TRANSACTION_VERSION
                ),
            )
            if receipt is not None:
                return json.loads(receipt.to_json())
            retries -= 1
            time.sleep(1)
        raise ValueError("Transaction receipt not found")

    def get_transaction(
        self,
        tx_digest: str,
        max_supported_transaction_version: Optional[int] = None,
        raise_on_try: bool = False,
    ) -> Optional[JSONLike]:
        """
        Get the transaction for a transaction digest.

        :param tx_digest: the digest associated to the transaction.
        :param max_supported_transaction_version: The max transaction version to return in responses.
        :param raise_on_try: whether the method will raise or log on error
        :return: the tx, if present
        """
        tx = self._try_get_transaction(
            tx_digest,
            max_supported_transaction_version=max_supported_transaction_version,
            raise_on_try=raise_on_try,
        )
        return tx

    @try_decorator("Error when attempting getting tx: {}", logger_method="debug")
    def _try_get_transaction(
        self,
        tx_digest: str,
        max_supported_transaction_version: Optional[int] = None,
        **_kwargs: Any,
    ) -> Optional[JSONLike]:
        """
        Get the transaction.

        :param tx_digest: the transaction digest.
        :param max_supported_transaction_version: The max transaction version to return in responses.
        :param _kwargs: the keyword arguments. Possible kwargs are:
            `raise_on_try`: bool flag specifying whether the method will raise or log on error (used by `try_decorator`)
        :return: the tx, if found
        """
        tx = self._api.get_transaction(
            Signature.from_string(tx_digest),
            max_supported_transaction_version=(
                max_supported_transaction_version
                or DEFAULT_MAX_SUPPORTED_TRANSACTION_VERSION
            ),
        )

        # pylint: disable=no-member
        return json.loads(tx.value.to_json())

    @staticmethod
    def create_default_account(
        from_address: str,
        new_account_address: str,
        lamports: int,
        space: int,
        program_id: Optional[str] = SYS_PROGRAM_ID,
    ):
        """
        Build a create account transaction.

        :param from_address: the sender public key
        :param new_account_address: the new account public key
        :param lamports: the amount of lamports to send
        :param space: the space to allocate
        :param program_id: the program id
        :return: the tx, if present
        """
        params = CreateAccountParams(
            PublicKey(from_address),
            PublicKey(new_account_address),
            lamports,
            space,
            PublicKey(program_id),
        )
        createAccountInstruction = create_account(params)
        txn = Transaction(fee_payer=from_address).add(createAccountInstruction)
        tx = txn._solders.to_json()  # pylint: disable=protected-access
        return json.loads(tx)

    def get_contract_instance(
        self, contract_interface: Dict[str, str], contract_address: Optional[str] = None
    ) -> Any:
        """
        Get the instance of a contract.

        :param contract_interface: the contract interface.
        :param contract_address: the contract address.
        :return: the contract instance
        """
        bytecode_path = None  # bytecode is not provided for the moment
        program_id = PublicKey.from_string(contract_address)
        idl = Idl.from_json(json.dumps(contract_interface["idl"]))
        pg = Program(idl, program_id)

        pg.provider.connection = self.api

        if bytecode_path is not None:
            # opening for [r]eading as [b]inary
            in_file = open(bytecode_path, "rb")
            bytecode = in_file.read()
        else:
            bytecode = None
        return {"program": pg, "bytecode": bytecode}

    def get_deploy_transaction(
        self,
        contract_interface: Dict[str, str],
        deployer_address: Address,
        raise_on_try: bool = False,
        **kwargs: Any,
    ) -> Optional[JSONLike]:
        """
        Get the transaction to deploy the smart contract.

        :param contract_interface: the contract interface.
        :param deployer_address: The address that will deploy the contract.
        :param raise_on_try: whether the method will raise or log on error
        :param kwargs: the keyword arguments.
        :returns tx: the transaction dictionary.
        """

        """
        if contract_interface["bytecode"] is None or contract_interface["program_keypair"] is None:
            raise ValueError("Bytecode or program_keypair is required")

        # check if solana cli is installed
        result = subprocess.run(
            ["solana --version"], capture_output=True, text=True, shell=True)
        if result.stderr != "":
            raise ValueError(result.stderr)

        # save keys in uint8 array temp
        value = struct.unpack('64B', payer_keypair.entity.secret_key)
        uint8_array = array.array('B', value)
        payer_uint8 = uint8_array.tolist()
        temp_dir_payer = Path(tempfile.mkdtemp())
        temp_file_payer = temp_dir_payer / "payer.json"
        temp_file_payer.write_text(str(payer_uint8))

        value = struct.unpack(
            '64B', contract_interface["program_keypair"].entity.secret_key)
        uint8_array = array.array('B', value)
        program_uint8 = uint8_array.tolist()
        temp_dir_program = Path(tempfile.mkdtemp())
        temp_file_program = temp_dir_program / "program.json"
        temp_file_program.write_text(str(program_uint8))

        t = SolanaCrypto(temp_file_payer)
        temp_dir_bytecode = Path(tempfile.mkdtemp())
        temp_file_bytecode = temp_dir_bytecode / "bytecode.so"
        temp_file_bytecode.write_bytes(contract_interface["bytecode"])

        cmd = f'''solana program deploy --url {DEFAULT_ADDRESS} -v --keypair {str(temp_file_payer)} --program-id {str(temp_file_program)} {str(temp_file_bytecode)}'''

        result = subprocess.run(
            [cmd], capture_output=True, text=True, shell=True)

        if result.stderr != "":
            raise ValueError(result.stderr)

        return result.stdout
        """
        raise NotImplementedError

    def contract_method_call(
        self,
        contract_instance: Any,
        method_name: str,
        **method_args: Any,
    ) -> Optional[JSONLike]:
        """
        Call a contract's method

        **TOBEIMPLEMENTED**

        :param contract_instance: the contract to use
        :param method_name: the contract method to call
        :param method_args: the contract call parameters

        # noqa: DAR202

        :return: the call result
        """
        raise NotImplementedError

    def build_transaction(  # pylint: disable=too-many-arguments
        self,
        contract_instance: Any,
        method_name: str,
        method_args: Optional[Dict[Any, Any]],
        tx_args: Optional[Dict[Any, Any]],
        raise_on_try: bool = False,
    ) -> Optional[JSONLike]:
        """Prepare a transaction

        :param contract_instance: the contract to use
        :param method_name: the contract method to call
        :param method_args: the contract parameters
        :param tx_args: the transaction parameters
        :param raise_on_try: whether the method will raise or log on error
        :return: the transaction
        """
        if method_args is None:
            raise ValueError("`method_args` can not be None")
        if method_args["data"] is None:
            raise ValueError("Data is required")
        if method_args["accounts"] is None:
            raise ValueError("Accounts are required")
        if tx_args.get("instruction_only", False):
            return self.build_instruction(
                contract_instance=contract_instance,
                method_name=method_name,
                data=method_args["data"],
                accounts=method_args["accounts"],
                remaining_accounts=method_args.get("remaining_accounts"),
            )
        tx = contract_instance.transaction[method_name](
            *method_args["data"],
            ctx=Context(
                accounts=method_args["accounts"],
                remaining_accounts=method_args.get("remaining_accounts"),
            ),
            payer=tx_args.get("payer"),
            blockhash=Hash.from_string(self.latest_hash),
        )
        return self.serialize_tx(tx=tx)

    def build_instruction(  # pylint: disable=too-many-arguments
        self,
        contract_instance: Program,
        method_name: str,
        data: List[Any],
        accounts: Dict[str, Pubkey],
        remaining_accounts: Optional[List[AccountMeta]] = None,
    ) -> JSONLike:
        """Prepare an instruction"""
        return self.serialize_ix(
            contract_instance.methods[method_name]
            .args(arguments=data)
            .accounts(accs=accounts)
            .remaining_accounts(accounts=remaining_accounts or [])
            .instruction()
        )

    def serialize_tx(
        self,
        tx: Union[
            Dict, SolanaTransaction, SoldersTransaction, SoldersVersionedTransaction
        ],
    ) -> Dict:
        """Serialize transaction to solders transaction compatible json object."""
        if isinstance(tx, Dict):
            return tx
        if isinstance(tx, SolanaTransaction):
            return json.loads(cast(SolanaTransaction, tx).to_solders().to_json())
        if isinstance(tx, BaseSolanaTransaction):
            return json.loads(tx.to_solders().to_json())
        if isinstance(tx, SoldersTransaction):
            return json.loads(cast(SoldersTransaction, tx).to_json())
        if isinstance(tx, SoldersVersionedTransaction):
            return json.loads(cast(SoldersVersionedTransaction, tx).to_json())
        raise ValueError(f"Unknown transction type found `{type(tx)}` ")

    def deserialize_tx(
        self,
        tx: Union[
            Dict, SolanaTransaction, SoldersTransaction, SoldersVersionedTransaction
        ],
    ) -> SolanaTransaction:
        """Deserialize transaction to a solana transaction object."""
        if isinstance(tx, SolanaTransaction):
            return cast(SolanaTransaction, tx)
        if isinstance(tx, SoldersTransaction):
            return SolanaTransaction.from_solders(txn=tx)
        if isinstance(tx, SoldersVersionedTransaction):
            return SolanaTransaction.from_solders(txn=tx.into_legacy_transaction())
        if not isinstance(tx, Dict):
            raise ValueError(f"Unknown transction type found `{type(tx)}`")

        if isinstance(tx["message"], list):
            _, tx["message"] = tx["message"]
            _, tx["signatures"] = tx["signatures"]
            return SolanaTransaction.from_solders(
                SoldersVersionedTransaction.populate(
                    message=MessageV0.from_json(json.dumps(tx["message"])),
                    signatures=[Signature.from_bytes(tx["signatures"])],
                )
            )

        # TODO: Safeguard for tx serialized to solders
        return SolanaTransaction.from_solders(
            SoldersTransaction.from_json(
                json.dumps(tx),
            )
        )

    def serialize_ix(self, ix: Instruction) -> Dict:
        """Serialize instruction."""
        return json.loads(ix.to_json())

    def deserialize_ix(self, ix: Dict) -> Instruction:
        """Deserialize instruction."""
        return Instruction.from_json(json.dumps(ix))

    def get_transaction_transfer_logs(  # pylint: disable=too-many-arguments,too-many-locals
        self,
        contract_instance: Any,
        tx_hash: str,
        max_supported_transaction_version: Optional[int] = None,
        target_address: Optional[str] = None,
    ) -> Optional[JSONLike]:
        """
        Get all transfer events derived from a transaction.

        :param contract_instance: contract instance
        :param tx_hash: the transaction hash
        :param max_supported_transaction_version: The max transaction version to return in responses.
        :param target_address: optional address to filter tranfer events to just those that affect it
        :return: the transfer logs
        """
        try:
            tx_receipt = self.get_transaction_receipt(
                tx_hash,
                max_supported_transaction_version=max_supported_transaction_version,
            )
            if tx_receipt is None:
                raise ValueError  # pragma: nocover

        except (  # pragma: nocover # pylint: disable=broad-except
            Exception,
            ValueError,
        ):
            return dict()

        keys = tx_receipt["transaction"]["message"]["accountKeys"]  # type: ignore
        if target_address:
            transfers = {
                "preBalances": [
                    {"address": keys[idx], "balance": balance}
                    for idx, balance in enumerate(tx_receipt["meta"]["preBalances"])  # type: ignore
                    if keys[idx] == target_address
                ],
                "postBalances": [
                    {"address": keys[idx], "balance": balance}
                    for idx, balance in enumerate(tx_receipt["meta"]["postBalances"])  # type: ignore
                    if keys[idx] == target_address
                ],
            }
        else:
            transfers = {
                "preBalances": [
                    {"address": keys[idx], "balance": balance}
                    for idx, balance in enumerate(tx_receipt["meta"]["preBalances"])  # type: ignore
                ],
                "postBalances": [
                    {"address": keys[idx], "balance": balance}
                    for idx, balance in enumerate(tx_receipt["meta"]["postBalances"])  # type: ignore
                ],
            }

        return transfers  # type: ignore  # actually ok

    def filter_event(
        self,
        event: Any,
        match_single: Dict[str, Any],
        match_any: Dict[str, Any],
        to_block: int,
        from_block: int,
        batch_size: int,
        max_retries: int,
        reduce_factor: float,
        timeout: int,
    ) -> Optional[JSONLike]:
        """Filter an event using batching to avoid RPC timeouts.

        :param event: the event to filter for.
        :param match_single: the filter parameters with value checking against the event abi. It allows for defining a single match value.
        :param match_any: the filter parameters with value checking against the event abi. It allows for defining multiple match values.
        :param to_block: the block to which to filter.
        :param from_block: the block from which to start filtering.
        :param batch_size: the blocks' batch size of the filtering.
        :param max_retries: the maximum number of retries.
        :param reduce_factor: the percentage by which the batch size is reduced in case of a timeout.
        :param timeout: a timeout in seconds to interrupt the operation in case the RPC request hangs.
        """
        raise NotImplementedError(  # pragma: nocover
            f"Custom events' filtering is not supported for the {self.identifier} plugin"
        )
